"""Preconfigured converters for tomllib."""

from base64 import b85decode, b85encode
from collections.abc import Set
from datetime import date, datetime
from enum import Enum
from operator import attrgetter
from typing import Any, TypeVar, Union

try:
    from tomllib import loads
except ImportError:
    from tomli import loads

try:
    from tomli_w import dumps
except ImportError:  # pragma: nocover
    dumps = None

from .._compat import is_mapping, is_subclass
from ..converters import BaseConverter, Converter
from ..fns import identity
from ..strategies import configure_union_passthrough
from . import validate_datetime, wrap

__all__ = ["TomllibConverter", "configure_converter", "make_converter"]

T = TypeVar("T")
_enum_value_getter = attrgetter("_value_")


class TomllibConverter(Converter):
    """A converter subclass specialized for tomllib."""

    if dumps is not None:

        def dumps(self, obj: Any, unstructure_as: Any = None, **kwargs: Any) -> str:
            return dumps(self.unstructure(obj, unstructure_as=unstructure_as), **kwargs)

    def loads(self, data: str, cl: type[T], **kwargs: Any) -> T:
        return self.structure(loads(data, **kwargs), cl)


def configure_converter(converter: BaseConverter):
    """
    Configure the converter for use with the tomllib library.

    * bytes are serialized as base85 strings
    * sets are serialized as lists
    * tuples are serializas as lists
    * mapping keys are coerced into strings when unstructuring
    * dates and datetimes are left for tomllib to handle
    """
    converter.register_structure_hook(bytes, lambda v, _: b85decode(v))
    converter.register_unstructure_hook(
        bytes, lambda v: (b85encode(v) if v else b"").decode("utf8")
    )

    @converter.register_unstructure_hook_factory(is_mapping)
    def gen_unstructure_mapping(cl: Any, unstructure_to=None):
        key_handler = str
        args = getattr(cl, "__args__", None)
        if args:
            if is_subclass(args[0], str):
                key_handler = _enum_value_getter if is_subclass(args[0], Enum) else None
            elif is_subclass(args[0], bytes):

                def key_handler(k: bytes):
                    return b85encode(k).decode("utf8")

        return converter.gen_unstructure_mapping(
            cl, unstructure_to=unstructure_to, key_handler=key_handler
        )

    converter.register_unstructure_hook(datetime, identity)
    converter.register_structure_hook(datetime, validate_datetime)
    converter.register_unstructure_hook(date, identity)
    converter.register_structure_hook(
        date, lambda v, _: v if isinstance(v, date) else date.fromisoformat(v)
    )
    configure_union_passthrough(Union[str, int, float, bool], converter)


@wrap(TomllibConverter)
def make_converter(*args: Any, **kwargs: Any) -> TomllibConverter:
    kwargs["unstruct_collection_overrides"] = {
        Set: list,
        tuple: list,
        **kwargs.get("unstruct_collection_overrides", {}),
    }
    res = TomllibConverter(*args, **kwargs)
    configure_converter(res)

    return res



import sys
from collections import Counter, deque
from collections.abc import Mapping as AbcMapping
from collections.abc import MutableMapping as AbcMutableMapping
from collections.abc import MutableSequence as AbcMutableSequence
from collections.abc import MutableSet as AbcMutableSet
from collections.abc import Sequence as AbcSequence
from collections.abc import Set as AbcSet
from dataclasses import MISSING, Field, is_dataclass
from dataclasses import fields as dataclass_fields
from functools import partial
from inspect import signature as _signature
from types import GenericAlias
from typing import (
    Annotated,
    Any,
    Deque,
    Dict,
    Final,
    FrozenSet,
    Generic,
    List,
    Literal,
    NewType,
    Optional,
    Protocol,
    Tuple,
    Union,
    _AnnotatedAlias,
    _GenericAlias,
    _SpecialGenericAlias,
    get_args,
    get_origin,
    get_type_hints,
)
from typing import Counter as TypingCounter
from typing import Mapping as TypingMapping
from typing import MutableMapping as TypingMutableMapping
from typing import MutableSequence as TypingMutableSequence
from typing import MutableSet as TypingMutableSet
from typing import Sequence as TypingSequence
from typing import Set as TypingSet

from attrs import NOTHING, Attribute, Factory, NothingType, resolve_types
from attrs import fields as attrs_fields
from attrs import fields_dict as attrs_fields_dict

__all__ = [
    "ANIES",
    "ExceptionGroup",
    "ExtensionsTypedDict",
    "TypeAlias",
    "adapted_fields",
    "fields_dict",
    "has",
    "is_typeddict",
]

try:
    from typing_extensions import TypedDict as ExtensionsTypedDict
except ImportError:  # pragma: no cover
    ExtensionsTypedDict = None

if sys.version_info >= (3, 11):
    from builtins import ExceptionGroup
else:
    from exceptiongroup import ExceptionGroup

try:
    from typing_extensions import is_typeddict as _is_typeddict
except ImportError:  # pragma: no cover
    assert sys.version_info >= (3, 10)
    from typing import is_typeddict as _is_typeddict

try:
    from typing_extensions import TypeAlias
except ImportError:  # pragma: no cover
    assert sys.version_info >= (3, 11)
    from typing import TypeAlias

LITERALS = {Literal}
try:
    from typing_extensions import Literal as teLiteral

    LITERALS.add(teLiteral)
except ImportError:  # pragma: no cover
    pass

# On some Python versions, `typing_extensions.Any` is different than
# `typing.Any`.
try:
    from typing_extensions import Any as teAny

    ANIES = frozenset([Any, teAny])
except ImportError:  # pragma: no cover
    ANIES = frozenset([Any])

NoneType = type(None)


def is_optional(typ: Any) -> bool:
    return is_union_type(typ) and NoneType in typ.__args__ and len(typ.__args__) == 2


def is_typeddict(cls: Any):
    """Thin wrapper around typing(_extensions).is_typeddict"""
    return _is_typeddict(getattr(cls, "__origin__", cls))


def has(cls):
    return hasattr(cls, "__attrs_attrs__") or hasattr(cls, "__dataclass_fields__")


def has_with_generic(cls):
    """Test whether the class if a normal or generic attrs or dataclass."""
    return has(cls) or has(get_origin(cls))


def fields(type):
    try:
        return type.__attrs_attrs__
    except AttributeError:
        return dataclass_fields(type)


def fields_dict(type) -> dict[str, Union[Attribute, Field]]:
    """Return the fields_dict for attrs and dataclasses."""
    if is_dataclass(type):
        return {f.name: f for f in dataclass_fields(type)}
    return attrs_fields_dict(type)


def adapted_fields(cl: type) -> list[Attribute]:
    """Return the attrs format of `fields()` for attrs and dataclasses.

    Resolves `attrs` stringified annotations, if present.
    """
    if is_dataclass(cl):
        attrs = dataclass_fields(cl)
        if any(isinstance(a.type, str) for a in attrs):
            # Do this conditionally in case `get_type_hints` fails, so
            # users can resolve on their own first.
            type_hints = get_type_hints(cl)
        else:
            type_hints = {}
        return [
            Attribute(
                attr.name,
                (
                    attr.default
                    if attr.default is not MISSING
                    else (
                        Factory(attr.default_factory)
                        if attr.default_factory is not MISSING
                        else NOTHING
                    )
                ),
                None,
                True,
                None,
                True,
                attr.init,
                True,
                type=type_hints.get(attr.name, attr.type),
                alias=attr.name,
                kw_only=getattr(attr, "kw_only", False),
            )
            for attr in attrs
        ]
    attribs = attrs_fields(cl)
    if any(isinstance(a.type, str) for a in attribs):
        # PEP 563 annotations - need to be resolved.
        resolve_types(cl)
        attribs = attrs_fields(cl)
    return attribs


def is_subclass(obj: type, bases) -> bool:
    """A safe version of issubclass (won't raise)."""
    try:
        return issubclass(obj, bases)
    except TypeError:
        return False


def is_hetero_tuple(type: Any) -> bool:
    origin = getattr(type, "__origin__", None)
    return origin is tuple and ... not in type.__args__


def is_protocol(type: Any) -> bool:
    return is_subclass(type, Protocol) and getattr(type, "_is_protocol", False)


def is_bare_final(type) -> bool:
    return type is Final


def get_final_base(type) -> Optional[type]:
    """Return the base of the Final annotation, if it is Final."""
    if type is Final:
        return Any
    if type.__class__ is _GenericAlias and type.__origin__ is Final:
        return type.__args__[0]
    return None


OriginAbstractSet = AbcSet
OriginMutableSet = AbcMutableSet

signature = partial(_signature, eval_str=True)


try:
    # Not present on 3.9.0, so we try carefully.
    from typing import _LiteralGenericAlias

    def is_literal(type: Any) -> bool:
        """Is this a literal?"""
        return type in LITERALS or (
            isinstance(
                type, (_GenericAlias, _LiteralGenericAlias, _SpecialGenericAlias)
            )
            and type.__origin__ in LITERALS
        )

except ImportError:  # pragma: no cover

    def is_literal(_) -> bool:
        return False


Set = AbcSet
MutableSet = AbcMutableSet
Sequence = AbcSequence
MutableSequence = AbcMutableSequence
MutableMapping = AbcMutableMapping
Mapping = AbcMapping
FrozenSetSubscriptable = frozenset
TupleSubscriptable = tuple


def is_annotated(type) -> bool:
    return getattr(type, "__class__", None) is _AnnotatedAlias


def is_tuple(type):
    return (
        type in (Tuple, tuple)
        or (type.__class__ is _GenericAlias and is_subclass(type.__origin__, Tuple))
        or (getattr(type, "__origin__", None) is tuple)
    )


if sys.version_info >= (3, 14):

    def is_union_type(obj):
        from types import UnionType  # noqa: PLC0415

        return obj is Union or isinstance(obj, UnionType)

    def get_newtype_base(typ: Any) -> Optional[type]:
        if typ is NewType or isinstance(typ, NewType):
            return typ.__supertype__
        return None

    from typing import NotRequired, Required

else:
    from typing import _UnionGenericAlias

    def is_union_type(obj):
        from types import UnionType  # noqa: PLC0415

        return (
            obj is Union
            or (isinstance(obj, _UnionGenericAlias) and obj.__origin__ is Union)
            or isinstance(obj, UnionType)
        )

    def get_newtype_base(typ: Any) -> Optional[type]:
        if typ is NewType or isinstance(typ, NewType):
            return typ.__supertype__
        return None

    if sys.version_info >= (3, 11):
        from typing import NotRequired, Required
    else:
        from typing_extensions import NotRequired, Required


def get_notrequired_base(type) -> Union[Any, NothingType]:
    if is_annotated(type):
        # Handle `Annotated[NotRequired[int]]`
        type = get_args(type)[0]
    if get_origin(type) in (NotRequired, Required):
        return get_args(type)[0]
    return NOTHING


def is_mutable_sequence(type: Any) -> bool:
    """A predicate function for mutable sequences.

    Matches lists, mutable sequences, and deques.
    """
    origin = getattr(type, "__origin__", None)
    return (
        type in (List, list, TypingMutableSequence, AbcMutableSequence, deque, Deque)
        or (
            type.__class__ is _GenericAlias
            and (
                ((origin is not tuple) and is_subclass(origin, TypingMutableSequence))
                or (origin is tuple and type.__args__[1] is ...)
            )
        )
        or (origin in (list, deque, AbcMutableSequence))
    )


def is_sequence(type: Any) -> bool:
    """A predicate function for sequences.

    Matches lists, sequences, mutable sequences, deques and homogenous
    tuples.
    """
    origin = getattr(type, "__origin__", None)
    return is_mutable_sequence(type) or (
        type in (TypingSequence, tuple, Tuple)
        or (
            type.__class__ is _GenericAlias
            and (
                ((origin is not tuple) and is_subclass(origin, TypingSequence))
                or (origin is tuple and type.__args__[1] is ...)
            )
        )
        or (origin is AbcSequence)
        or (origin is tuple and type.__args__[1] is ...)
    )


def is_deque(type):
    return (
        type in (deque, Deque)
        or (type.__class__ is _GenericAlias and is_subclass(type.__origin__, deque))
        or (getattr(type, "__origin__", None) is deque)
    )


def is_mutable_set(type: Any) -> bool:
    """A predicate function for (mutable) sets.

    Matches built-in sets and sets from the typing module.
    """
    return (
        type in (TypingSet, TypingMutableSet, set)
        or (
            type.__class__ is _GenericAlias
            and is_subclass(type.__origin__, TypingMutableSet)
        )
        or (getattr(type, "__origin__", None) in (set, AbcMutableSet, AbcSet))
    )


def is_frozenset(type: Any) -> bool:
    """A predicate function for frozensets.

    Matches built-in frozensets and frozensets from the typing module.
    """
    return (
        type in (FrozenSet, frozenset)
        or (type.__class__ is _GenericAlias and is_subclass(type.__origin__, FrozenSet))
        or (getattr(type, "__origin__", None) is frozenset)
    )


def is_bare(type):
    return isinstance(type, _SpecialGenericAlias) or (
        not hasattr(type, "__origin__") and not hasattr(type, "__args__")
    )


def is_mapping(type: Any) -> bool:
    """A predicate function for mappings."""
    return (
        type in (dict, Dict, TypingMapping, TypingMutableMapping, AbcMutableMapping)
        or (
            type.__class__ is _GenericAlias
            and is_subclass(type.__origin__, TypingMapping)
        )
        or is_subclass(
            getattr(type, "__origin__", type), (dict, AbcMutableMapping, AbcMapping)
        )
    )


def is_counter(type):
    return (
        type in (Counter, TypingCounter) or getattr(type, "__origin__", None) is Counter
    )


def is_generic(type) -> bool:
    """Whether `type` is a generic type."""
    # Inheriting from protocol will inject `Generic` into the MRO
    # without `__orig_bases__`.
    return (
        isinstance(type, (_GenericAlias, GenericAlias))
        or (is_subclass(type, Generic) and hasattr(type, "__orig_bases__"))
        or type.__class__ is Union  # On 3.14, unions are no longer typing._GenericAlias
    )


def copy_with(type, args):
    """Replace a generic type's arguments."""
    if is_annotated(type):
        # typing.Annotated requires a special case.
        return Annotated[args]
    if isinstance(args, tuple) and len(args) == 1:
        # Some annotations can't handle 1-tuples.
        args = args[0]
    return type.__origin__[args]


def get_full_type_hints(obj, globalns=None, localns=None):
    return get_type_hints(obj, globalns, localns, include_extras=True)


def is_generic_attrs(type) -> bool:
    """Return True for both specialized (A[int]) and unspecialized (A) generics."""
    return is_generic(type) and has(type.__origin__)



import sys

if sys.version_info <= (3, 13):

    def subclasses(cls: type) -> list[type]:
        """A proxy for `cls.__subclasses__()` on older Pythons."""
        return cls.__subclasses__()

else:

    def subclasses(cls: type) -> list[type]:
        """A helper for getting subclasses of a class.

        Filters out duplicate subclasses of slot dataclasses and attrs classes.
        """
        return [
            cl
            for cl in cls.__subclasses__()
            if (
                not (
                    "__slots__" not in cl.__dict__
                    and hasattr(cls, "__dataclass_params__")
                    and cls.__dataclass_params__.slots
                )
                and not hasattr(cls, "__attrs_base_of_slotted__")
            )
        ]



from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .converters import BaseConverter


def enum_unstructure_factory(
    type: type[Enum], converter: "BaseConverter"
) -> Callable[[Enum], Any]:
    """A factory for generating enum unstructure hooks.

    If the enum is a typed enum (has `_value_`), we use the underlying value's hook.
    Otherwise, we use the value directly.
    """
    if "_value_" in type.__annotations__:
        return lambda e: converter.unstructure(e.value)

    return lambda e: e.value


def enum_structure_factory(
    type: type[Enum], converter: "BaseConverter"
) -> Callable[[Any, type[Enum]], Enum]:
    """A factory for generating enum structure hooks.

    If the enum is a typed enum (has `_value_`), we structure the value first.
    Otherwise, we use the value directly.
    """
    if "_value_" in type.__annotations__:
        val_type = type.__annotations__["_value_"]
        val_hook = converter.get_structure_hook(val_type)
        return lambda v, _: type(val_hook(v, val_type))

    return lambda v, _: type(v)



from collections.abc import Sequence
from typing import Any, Optional, Union

from typing_extensions import Self

from cattrs._compat import ExceptionGroup


class StructureHandlerNotFoundError(Exception):
    """
    Error raised when structuring cannot find a handler for converting inputs into
    :attr:`type_`.
    """

    def __init__(self, message: str, type_: type) -> None:
        super().__init__(message, type_)
        self.message = message
        self.type_ = type_

    def __str__(self) -> str:
        return self.message


class BaseValidationError(ExceptionGroup):
    cl: type

    def __new__(cls, message: str, excs: Sequence[Exception], cl: type) -> Self:
        obj = super().__new__(cls, message, excs)
        obj.cl = cl
        return obj

    def derive(self, excs: Sequence[Exception]) -> Self:
        return self.__class__(self.message, excs, self.cl)


class IterableValidationNote(str):
    """Attached as a note to an exception when an iterable element fails structuring."""

    index: Union[int, str]  # Ints for list indices, strs for dict keys
    type: Any

    def __new__(cls, string: str, index: Union[int, str], type: Any) -> Self:
        instance = str.__new__(cls, string)
        instance.index = index
        instance.type = type
        return instance

    def __getnewargs__(self) -> tuple[str, Union[int, str], Any]:
        return (str(self), self.index, self.type)


class IterableValidationError(BaseValidationError):
    """Raised when structuring an iterable."""

    def group_exceptions(
        self,
    ) -> tuple[list[tuple[Exception, IterableValidationNote]], list[Exception]]:
        """Split the exceptions into two groups: with and without validation notes."""
        excs_with_notes = []
        other_excs = []
        for subexc in self.exceptions:
            if hasattr(subexc, "__notes__"):
                for note in subexc.__notes__:
                    if note.__class__ is IterableValidationNote:
                        excs_with_notes.append((subexc, note))
                        break
                else:
                    other_excs.append(subexc)
            else:
                other_excs.append(subexc)

        return excs_with_notes, other_excs


class AttributeValidationNote(str):
    """Attached as a note to an exception when an attribute fails structuring."""

    name: str
    type: Any

    def __new__(cls, string: str, name: str, type: Any) -> Self:
        instance = str.__new__(cls, string)
        instance.name = name
        instance.type = type
        return instance

    def __getnewargs__(self) -> tuple[str, str, Any]:
        return (str(self), self.name, self.type)


class ClassValidationError(BaseValidationError):
    """Raised when validating a class if any attributes are invalid."""

    def group_exceptions(
        self,
    ) -> tuple[list[tuple[Exception, AttributeValidationNote]], list[Exception]]:
        """Split the exceptions into two groups: with and without validation notes."""
        excs_with_notes = []
        other_excs = []
        for subexc in self.exceptions:
            if hasattr(subexc, "__notes__"):
                for note in subexc.__notes__:
                    if note.__class__ is AttributeValidationNote:
                        excs_with_notes.append((subexc, note))
                        break
                else:
                    other_excs.append(subexc)
            else:
                other_excs.append(subexc)

        return excs_with_notes, other_excs


class ForbiddenExtraKeysError(Exception):
    """
    Raised when `forbid_extra_keys` is activated and such extra keys are detected
    during structuring.

    The attribute `extra_fields` is a sequence of those extra keys, which were the
    cause of this error, and `cl` is the class which was structured with those extra
    keys.
    """

    def __init__(
        self, message: Optional[str], cl: type, extra_fields: set[str]
    ) -> None:
        self.message = message
        self.cl = cl
        self.extra_fields = extra_fields

        super().__init__(message, cl, extra_fields)

    def __str__(self) -> str:
        return (
            self.message
            or f"Extra fields in constructor for {self.cl.__name__}: "
            f"{', '.join(sorted(self.extra_fields))}"
        )
